use crate::lexer::Token;
use chumsky::{
    input::{Stream, ValueInput},
    prelude::*,
};

type Span = SimpleSpan;
type Spanned<T> = (T, Span);

#[derive(Debug, PartialEq)]
pub enum Expr {
    Number(f64),
    String(String),
    Bool(bool),
    Undefined,
    Some(Box<Expr>),
    None,
    Ok(Box<Expr>),
    Err(Box<Expr>),
}

/// parse primary tokens, like numbers, strings or booleans
pub fn primary<'tokens, 'src: 'tokens, I>()
-> impl Parser<'tokens, I, Expr, extra::Err<Rich<'tokens, Token<'src>>>>
where
    I: ValueInput<'tokens, Token = Token<'src>, Span = SimpleSpan>,
{
    select! {
        Token::Number(n) => Expr::Number(n),
        Token::String(s) => Expr::String(s),
        Token::Undefined => Expr::Undefined,
        Token::None => Expr::None,
        Token::Bool(b) => Expr::Bool(b),
    }
}

fn parenthesized<'tokens, 'src: 'tokens, I>(
    inner: impl Parser<'tokens, I, Expr, extra::Err<Rich<'tokens, Token<'src>>>>,
) -> impl Parser<'tokens, I, Expr, extra::Err<Rich<'tokens, Token<'src>>>>
where
    I: ValueInput<'tokens, Token = Token<'src>, Span = SimpleSpan>,
{
    just(Token::ParenOpen)
        .ignore_then(inner)
        .then_ignore(just(Token::ParenClose))
}

fn constructor<'tokens, 'src: 'tokens, I>(
    keyword: Token<'src>,
    constructor_fn: impl Fn(Expr) -> Expr,
) -> impl Parser<'tokens, I, Expr, extra::Err<Rich<'tokens, Token<'src>>>>
where
    I: ValueInput<'tokens, Token = Token<'src>, Span = SimpleSpan>,
{
    just(keyword)
        .ignore_then(parenthesized(primary()))
        .map(constructor_fn)
}

pub fn constructors<'tokens, 'src: 'tokens, I>()
-> impl Parser<'tokens, I, Expr, extra::Err<Rich<'tokens, Token<'src>>>>
where
    I: ValueInput<'tokens, Token = Token<'src>, Span = SimpleSpan>,
{
    constructor(Token::Some, |inner| Expr::Some(Box::new(inner)))
        .or(constructor(Token::Ok, |inner| Expr::Ok(Box::new(inner))))
        .or(constructor(Token::Err, |inner| Expr::Err(Box::new(inner))))
}

#[cfg(test)]
mod tests {
    use super::*;
    use logos::Logos;

    #[test]
    fn test_parse_single_number() {
        let tokens: Vec<_> = Token::lexer("42").map(|tok| tok.unwrap()).collect();
        let result = primary().parse(Stream::from_iter(tokens)).into_result();
        assert_eq!(result, Ok(Expr::Number(42.0)));
    }
    #[test]
    fn test_parse_single_string() {
        let tokens: Vec<_> = Token::lexer("\"Foo\"").map(|tok| tok.unwrap()).collect();
        let result = primary().parse(Stream::from_iter(tokens)).into_result();
        assert_eq!(result, Ok(Expr::String("Foo".to_owned())));
    }
    #[test]
    fn test_constructors() {
        let tokens_some: Vec<_> = Token::lexer("Some(23)").map(|tok| tok.unwrap()).collect();
        let result_some = constructors()
            .parse(Stream::from_iter(tokens_some))
            .into_result();
        assert_eq!(result_some, Ok(Expr::Some(Box::new(Expr::Number(23.0)))));

        let tokens_ok: Vec<_> = Token::lexer("Ok(\"foo\")")
            .map(|tok| tok.unwrap())
            .collect();
        let result_ok = constructors()
            .parse(Stream::from_iter(tokens_ok))
            .into_result();
        assert_eq!(
            result_ok,
            Ok(Expr::Ok(Box::new(Expr::String("foo".to_owned()))))
        );

        let tokens_err: Vec<_> = Token::lexer("Err(None)").map(|tok| tok.unwrap()).collect();
        let result_err = constructors()
            .parse(Stream::from_iter(tokens_err))
            .into_result();
        assert_eq!(result_err, Ok(Expr::Err(Box::new(Expr::None))));
    }
}
