use crate::parser;
use chumsky;
use std::collections::HashMap;
use swc_common::{BytePos, DUMMY_SP, Span, SyntaxContext};
use swc_ecma_ast::{
    AssignExpr, BinaryOp, BindingIdent, BlockStmt, Decl, Expr, FnDecl, Function, Ident, Module,
    ModuleItem, Param, Pat, Stmt,
};

pub trait ToSWC<T> {
    fn to_swc(&self) -> T;
}

impl ToSWC<Span> for chumsky::span::SimpleSpan {
    fn to_swc(&self) -> Span {
        Span::new(BytePos(self.start as u32), BytePos(self.end as u32))
    }
}

impl ToSWC<BinaryOp> for parser::BinaryOp {
    fn to_swc(&self) -> BinaryOp {
        match self {
            parser::BinaryOp::Add => BinaryOp::Add,
            parser::BinaryOp::Sub => BinaryOp::Sub,
            parser::BinaryOp::Mul => BinaryOp::Mul,
            parser::BinaryOp::Div => BinaryOp::Div,
            parser::BinaryOp::Eq => BinaryOp::EqEqEq,
            parser::BinaryOp::NotEq => BinaryOp::NotEqEq,
            // TODO: implement all members of BinaryOp
        }
    }
}

impl ToSWC<Function> for parser::Func<'_> {
    fn to_swc(&self) -> Function {
        Function {
            params: self
                .args
                .iter()
                .map(|name| Param {
                    span: DUMMY_SP,
                    decorators: vec![],
                    pat: Pat::Ident(BindingIdent {
                        id: name.to_string().into(),
                        type_ann: None,
                    }),
                })
                .collect(),
            decorators: vec![],
            span: self.span.to_swc(),
            body: Some(BlockStmt {
                span: DUMMY_SP,
                ctxt: SyntaxContext::empty(),
                stmts: vec![], //TODO!
            }),
            is_generator: false,
            is_async: false,
            type_params: None,
            return_type: None,
            ctxt: SyntaxContext::empty(),
        }
    }
}

impl ToSWC<Expr> for parser::Expr<'_> {
    fn to_swc(&self) -> Expr {
        match self {
            parser::Expr::Var(name, val, body) => Expr::Assign(AssignExpr {
                span: Span::new(BytePos(self.1.1.start as u32), BytePos(self.2.1.end as u32)),
            }),
        }
    }
}

pub struct JsTransformer;

impl<'src> JsTransformer {
    pub fn new() -> Self {
        Self {}
    }
    pub fn transform(
        &self,
        source_ast: HashMap<&'src str, parser::Func<'src>>,
        span: chumsky::span::SimpleSpan,
    ) -> Module {
        Module {
            span: span.to_swc(),
            body: source_ast
                .into_iter()
                .map(|(name, func)| self.transform_func_stmt(name, func))
                .collect(),
            shebang: None,
        }
    }
    pub fn transform_func_stmt(&self, ident: &str, func: parser::Func<'_>) -> ModuleItem {
        ModuleItem::Stmt(Stmt::Decl(Decl::Fn(FnDecl {
            ident: Ident::new(ident.into(), DUMMY_SP, SyntaxContext::empty()),
            declare: false,
            function: Box::new(func.to_swc()),
        })))
    }

    // pub fn transform_expr(&self, expr: parser::Expr) -> Stmt {
    //     match expr {
    //         Expr::Var => Stmt::Decl(Decl::Var(())),
    //     }
    // }
}
